/*
 * Copyright (C) 2022 Jonathan Certes
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>
#include <tcl.h>

/**
 * \brief
 *  When calling functions to be called when a Tcl procedure is invoked or
 *  deleted, custom client data can be transmitted.
 */
typedef struct {
  Tcl_Interp* theInterp;    //!< pointer to the Tcl interpreter

  int myVariable;           //!< a C variable we make available in Tcl
  char myVariableName[11];  //!< name of the Tcl variable

  Tcl_ObjCmdProc * myCommandHandler;  //!< a C function we make available in Tcl
  char myCommandName[10];             //!< name of the Tcl procedure
} theClientDataStruct;

/******************************************************************************/

/**
 * \brief
 *  The C function that is called when procedure \ref
 *  theClientDataStruct.myCommandName is called by the interpreter.
 *
 * \details
 *  Its prototype matches a <tt> Tcl_ObjCmdProc </tt> as described in <tt> man
 *  3tcl Tcl_CreateObjCommand </tt>.
 *
 *  This function increments theClientDataStruct.myVariable.
 *
 * \returns either TCL_OK, TCL_ERROR, TCL_RETURN, TCL_BREAK, or TCL_CONTINUE
 */
static int incrMyVar(
  ClientData clientData,  //!< pointer to the client data
  Tcl_Interp *interp,     //!< pointer to the Tcl interpreter
  int objc,   //!< number of argument values (including the command name)
  Tcl_Obj *const objv[]   //!< values of the arguments
) {
  theClientDataStruct * theClientData = (theClientDataStruct *)( clientData );

  if ( objc != 1 ) {
    Tcl_WrongNumArgs(interp, 1, objv, "");
    return TCL_ERROR;
  }

  theClientData->myVariable = theClientData->myVariable + 1;
  // so that Tcl trace can see it:
  Tcl_UpdateLinkedVar(interp, theClientData->myVariableName);
  return TCL_OK;
}

/******************************************************************************/

/**
 * \brief
 *  The C function that is called when procedure \ref
 *  theClientDataStruct.myCommandName is deleted by the interpreter. It
 *  implements command-specific cleanup and is optional.
 *
 * \details
 *  Its prototype matches a <tt> Tcl_CmdDeleteProc </tt> as described in <tt>
 *  man 3tcl Tcl_CreateObjCommand </tt>.
 *
 *  Unlinks the shared variable and frees the memory.
 */
static void deleteMyCommand(
  ClientData clientData //!< pointer to the client data
) {
  theClientDataStruct * theClientData = (theClientDataStruct *)( clientData );

  Tcl_UnlinkVar(theClientData->theInterp, theClientData->myVariableName);
  Tcl_Release(theClientData);
  Tcl_Free(clientData);
}

/******************************************************************************/

/**
 * \brief
 *  Function that is called when the shared library is loaded.
 *
 * \details
 *  Makes one shared variable and one shared command available in Tcl.
 *
 * \returns either TCL_OK, TCL_ERROR, TCL_RETURN, TCL_BREAK, or TCL_CONTINUE
 */
int Example_Init(
  Tcl_Interp *interp  //!< pointer to the Tcl interpreter
) {
  theClientDataStruct * theClientData;

  if ( Tcl_InitStubs(interp, "8.6", 0) == NULL ) {
    return TCL_ERROR;
  }

  // initializes the structure:
  theClientData = (theClientDataStruct *)(
                                      Tcl_Alloc(sizeof(theClientDataStruct)) );
  theClientData->theInterp  = interp;
  theClientData->myVariable = 0;
  strncpy(theClientData->myVariableName, "myVariable",
                                        sizeof(theClientData->myVariableName));
  theClientData->myCommandHandler = incrMyVar;
  strncpy(theClientData->myCommandName, "myCommand",
                                        sizeof(theClientData->myCommandName));
  Tcl_Preserve(theClientData);  // makes sure data is kept until Tcl_Release()

  // makes a C variable available in Tcl:
  Tcl_LinkVar(  interp,
                theClientData->myVariableName,
                (char *)( &theClientData->myVariable ),
                TCL_LINK_INT
              );

  // makes a C function available in Tcl:
  Tcl_CreateObjCommand( interp,
                        theClientData->myCommandName,
                        theClientData->myCommandHandler,
                        theClientData,
                        deleteMyCommand // optional command-specific cleanup
                      );

  // the package name must match the prefix of the name of this function:
  Tcl_PkgProvide(interp, "example", "1.0");

  return TCL_OK;
}

